#ifndef _itkImageToVesselnessMeasureImageFunction_h
#define _itkImageToVesselnessMeasureImageFunction_h

#include "itkImageFunction.h"
#include "itkNumericTraits.h"
#include "itkImageToImageFilter.h"
#include "itkHessianRecursiveGaussianImageFilter.h"

namespace itk
{

/**
 * \class ImageToVesselnessMeasureImageFunction
 *
 * 
 * \ingroup ImageFunctions
 */
template <class TInputImage, class TOutputImage, typename TCoordRep = float >
class ITK_EXPORT ImageToVesselnessMeasureImageFunction :
public ImageFunction< TInputImage, ITK_TYPENAME TOutputImage::PixelType, TCoordRep >
{
public:
  /** Standard class typedefs. */
  typedef ImageToVesselnessMeasureImageFunction Self;
  typedef ImageFunction< TInputImage, ITK_TYPENAME TOutputImage::PixelType, TCoordRep > Superclass;
  typedef SmartPointer<Self> Pointer;
  typedef SmartPointer<const Self>  ConstPointer;
  
  /** Run-time type information (and related methods). */
  itkTypeMacro(ImageToVesselnessMeasureImageFunction, ImageFunction);

  /** Method for creation through the object factory. */
  itkNewMacro(Self);

  /** InputImageType typedef support. */
  typedef TInputImage InputImageType;

  /** OutputType typdef support. */
  typedef typename Superclass::OutputType OutputType;

  /** Index typedef support. */
  typedef typename Superclass::IndexType IndexType;
  
  /** ContinuousIndex typedef support. */
  typedef typename Superclass::ContinuousIndexType ContinuousIndexType;

  /** Point typedef support. */
  typedef typename Superclass::PointType PointType;

	typedef itk::HessianRecursiveGaussianImageFilter< InputImageType > HessianRecursiveGaussianFilterType;
	typedef itk::FixedArray< double, ::itk::GetImageDimension<InputImageType>::ImageDimension >  EigenValueArrayType;
	typedef typename HessianRecursiveGaussianFilterType::OutputPixelType::MatrixType EigenVectorsMatrixType;
	typedef itk::SymmetricEigenAnalysis< 
								typename HessianRecursiveGaussianFilterType::OutputPixelType, 
								EigenValueArrayType,
								EigenVectorsMatrixType >			EigenCalculatorType;

	/** Set/Get macros for alpha_1 */
  itkSetMacro(Alpha1, double);
  itkGetMacro(Alpha1, double);
  
  /** Set/Get macros for alpha_2 */
  itkSetMacro(Alpha2, double);
  itkGetMacro(Alpha2, double);

	/** Dimension of the underlying image. */
  itkStaticConstMacro(ImageDimension, unsigned int,
                      InputImageType::ImageDimension);

  /** Evalulate the function at specified index */
  virtual OutputType EvaluateAtIndex( const IndexType& index ) const;
  
  /** Evaluate the function at non-integer positions */
  virtual OutputType Evaluate( const PointType& point ) const
    { 
      IndexType index;
      this->ConvertPointToNearestIndex( point, index );
      return this->EvaluateAtIndex( index ); 
    }
  virtual OutputType EvaluateAtContinuousIndex( 
    const ContinuousIndexType& cindex ) const
    { 
      IndexType index;
      this->ConvertContinuousIndexToNearestIndex( cindex, index );
      return this->EvaluateAtIndex( index ) ; 
    }

	void SetSigma( double s )
	{
		m_SigmaList.clear();
		m_SigmaList.push_back(s);
		m_HessianRecursiveGaussianFilter->SetSigma( s );
		this->Modified();
	}

	double GetSigma( void ) const
	{
		return m_SigmaList.back();
	}

	void AddSigma( double s )
	{
		m_SigmaList.push_back( s );
		m_HessianRecursiveGaussianFilter->SetSigma( s );
		this->Modified();
	}	

  itkSetMacro(MultiScale, bool);
  itkGetMacro(MultiScale, bool);

	void SetNormalizeAcrossScale( bool flag )
	{
		m_HessianRecursiveGaussianFilter->SetNormalizeAcrossScale( flag );
	}

	bool GetNormalizeAcrossScale( void ) const
	{
		return m_HessianRecursiveGaussianFilter->GetNormalizeAcrossScale();
	}

	virtual void SetInputImage(	const InputImageType*  ptr );
	
protected:
  ImageToVesselnessMeasureImageFunction();
  ~ImageToVesselnessMeasureImageFunction(){};
  void PrintSelf(std::ostream& os, Indent indent) const;
	
private:
  ImageToVesselnessMeasureImageFunction( const Self& ); //purposely not implemented
  void operator=( const Self& ); //purposely not implemented

	double m_Alpha1;
  double m_Alpha2;
	typename InputImageType::Pointer m_ShallowInput;
	mutable typename TOutputImage::RegionType m_ShallowOutputRegion;
	EigenCalculatorType m_EigenCalculator;
	typename HessianRecursiveGaussianFilterType::Pointer m_HessianRecursiveGaussianFilter;
	mutable typename EigenCalculatorType::VectorType m_EigenValues;
	mutable typename EigenCalculatorType::EigenMatrixType m_EigenVectors;
	std::vector< double > m_SigmaList;
	bool m_MultiScale;
	
};

} // namespace itk

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkImageToVesselnessMeasureImageFunction.txx"
#endif

#endif

