#ifndef _itkImageToVesselnessMeasureImageFunction_txx
#define _itkImageToVesselnessMeasureImageFunction_txx
#include "itkImageToVesselnessMeasureImageFunction.h"

#include "itkNumericTraits.h"
#include <iostream>

namespace itk
{

/**
   * Constructor
   */
template <class TInputImage, class TOutputImage, typename TCoordRep >
ImageToVesselnessMeasureImageFunction<TInputImage,TOutputImage, TCoordRep>
::ImageToVesselnessMeasureImageFunction()
: m_MultiScale( false )
{
  m_ShallowInput = InputImageType::New();
	typename TOutputImage::SizeType size; 
	size.Fill(1);
	m_ShallowOutputRegion.SetSize(size);

	// set default values of alphas
	m_Alpha1 = 0.5;
  m_Alpha2 = 2.0;

	// set Eigen calculator params
  m_EigenCalculator.SetDimension(ImageDimension);

	// Create filter that computes hessian matrices
	m_HessianRecursiveGaussianFilter = HessianRecursiveGaussianFilterType::New();
	m_HessianRecursiveGaussianFilter->SetNormalizeAcrossScale( true );

	this->SetSigma( 1.0 );
}


/**
   *
   */
template <class TInputImage, class TOutputImage, typename TCoordRep >
void
ImageToVesselnessMeasureImageFunction<TInputImage,TOutputImage, TCoordRep>
::PrintSelf(std::ostream& os, Indent indent) const
{
  this->Superclass::PrintSelf(os,indent);

	os << indent << "Internal Hessian: " << m_HessianRecursiveGaussianFilter << std::endl;
	os << indent << "Alpha1: " << m_Alpha1 << std::endl;
	os << indent << "Alpha2: " << m_Alpha2 << std::endl;	
}

template <class TInputImage, class TOutputImage, typename TCoordRep >
void 
ImageToVesselnessMeasureImageFunction<TInputImage,TOutputImage, TCoordRep>
::SetInputImage(	const InputImageType*  ptr )
{
	// Create/Update shallow input and pass it to internal filter
	if( !ptr  )
	{
		itkDebugMacro(<< "ImageToVesselnessMeasureImageFunction::SetInputImage Error input invalid.\n");
		return;
	}

	// Set the shallowInput
	// copy regions
	m_ShallowInput->SetLargestPossibleRegion( ptr->GetLargestPossibleRegion() );
	m_ShallowInput->SetBufferedRegion( ptr->GetBufferedRegion() );
	m_ShallowInput->SetRequestedRegion( ptr->GetRequestedRegion() );

	// copy PixelContainer so content is shared across inputs 
	// and no allocation is required
	m_ShallowInput->SetPixelContainer( const_cast< InputImageType::PixelContainer* >( ptr->GetPixelContainer() ) );

	// Assign as input to internal pipeline
	m_HessianRecursiveGaussianFilter->SetInput( m_ShallowInput );

	// Finally call the superclass SetInputImage function
	Superclass::SetInputImage( ptr );
}

/**
 *
 */
template <class TInputImage, class TOutputImage, typename TCoordRep >
typename ImageToVesselnessMeasureImageFunction<TInputImage,TOutputImage, TCoordRep>::OutputType
ImageToVesselnessMeasureImageFunction<TInputImage,TOutputImage, TCoordRep>
::EvaluateAtIndex(const IndexType& index) const
{
	if ( !this->IsInsideBuffer( index ) )
  {
		return itk::NumericTraits< OutputType >::min();
  }

	m_ShallowOutputRegion.SetIndex( index );

	// Run mini-pipeline
	m_HessianRecursiveGaussianFilter->GetOutput()->SetRequestedRegion( m_ShallowOutputRegion );

	std::vector< double >::const_iterator itBegin = m_SigmaList.begin();
	std::vector< double >::const_iterator itEnd;
	if( m_MultiScale )
	{
		itEnd = m_SigmaList.end();
	}
	else
	{
		itEnd = itBegin;
		++itEnd;
	}
	
	std::vector< double >::const_iterator itSigma;
	double lineMeasureMax = itk::NumericTraits< OutputType >::min();

	for( itSigma = itBegin; itSigma != itEnd; ++itSigma )
	{

		// Set Sigma
		m_HessianRecursiveGaussianFilter->SetSigma( *itSigma );

		// Update
		m_HessianRecursiveGaussianFilter->Update();

		// Get hessian mat
		const typename EigenCalculatorType::MatrixType& hessianMat = m_HessianRecursiveGaussianFilter->GetOutput()->GetPixel( index );

		// Perform eigen analysis
		m_EigenCalculator.ComputeEigenValuesAndVectors( hessianMat, m_EigenValues, m_EigenVectors );

		// Compute vesselness
		double normalizeValue = vnl_math_min( -1.0 * m_EigenValues[1], -1.0 * m_EigenValues[0] );

		// Similarity measure to a line structure
		double lineMeasure = 0.0;

		if( normalizeValue > 0 )
		{

			if( m_EigenValues[2] <= 0 )
			{
				lineMeasure = 
					exp( -0.5 * vnl_math_sqr( m_EigenValues[2] / (m_Alpha1 * normalizeValue)));
			}
			else
			{
				lineMeasure = 
					exp( -0.5 * vnl_math_sqr( m_EigenValues[2] / (m_Alpha2 * normalizeValue)));
			}      
			lineMeasure *= normalizeValue;				
		}

		lineMeasureMax = lineMeasure > lineMeasureMax ? lineMeasure : lineMeasureMax;

	}

	return static_cast< OutputType >( lineMeasureMax );
}


} // namespace itk

#endif
